"""Script to allow argdown parse_args to markdown conversion"""

import inspect
from lbsntransform import BaseConfig
from lbsntransform import __version__


def extract_argscode():
    """Extracts command line args code to separate file

    Preparation step for processing with argdown
    """
    # open file to output source code
    source_file = open("parse_args.py", "w")
    # extract source code of parse_args
    parse_args_source = inspect.getsource(BaseConfig.get_arg_parser)
    # remove first three lines
    for __ in range(0, 3):
        parse_args_source = parse_args_source[parse_args_source.index("\n") + 1 :]
    # insert at beginning
    parse_args_source = (
        f'parser = argparse.ArgumentParser(prog="lbsntransform")\n{parse_args_source}'
    )
    # unindent all other lines
    parse_args_source = parse_args_source.lstrip().replace("\n        ", "\n")
    # replace version string
    parse_args_source = parse_args_source.replace(
        "{__version__}", f"{__version__}"
    )
    # replace package name
    parse_args_source = parse_args_source.replace(
        "usage: argdown", "usage: lbsntransform"
    )
    # write argdown and argparse imports first
    source_file.write("import argparse\n")
    source_file.write("import argdown\n")
    source_file.write("from pathlib import Path\n")
    # fix argparse name
    parse_args_source = parse_args_source.replace(
        "ArgumentParser()", 'ArgumentParser(prog="lbsntransform")'
    )
    # replace method leftover
    parse_args_source = parse_args_source.replace(
        "return parser", "args = parser.parse_args()"
    )
    # replace two empty spaces on str line ends with non-breaking spaces,
    # protecting line breaks during python-markdown conversion
    parse_args_source = parse_args_source.replace("'  '\n", "'&nbsp;&nbsp;<br>'\n")
    parse_args_source = parse_args_source.replace("  '\n", "&nbsp;&nbsp;<br>'\n")
    # protected indent
    parse_args_source = parse_args_source.replace("'    ", "'&nbsp;&nbsp;&nbsp;&nbsp;")
    # write prepared source code
    source_file.write(parse_args_source)
    source_file.close()


# run script
extract_argscode()



# -*- coding: utf-8 -*-

"""
Module for mapping Instagram (json) to common LBSN Structure.
"""

import json

# pylint: disable=no-member
import logging
import re
from typing import Any, Dict

import lbsnstructure as lbsn

from lbsntransform.tools.helper_functions import HelperFunctions as HF

MAPPING_ID = 14


class importer:
    """Provides mapping function from Instagram endpoints to
    protobuf lbsnstructure
    """

    ORIGIN_NAME = "Instagram"
    ORIGIN_ID = 1

    def __init__(
        self,
        ignore_non_geotagged=False,
        min_geoaccuracy=None,
        **_
    ):
        # Create the OriginID globally
        # this OriginID is required for all CompositeKeys
        origin = lbsn.Origin()
        origin.origin_id = lbsn.Origin.INSTAGRAM
        self.origin = origin
        # this is where all the data will be stored
        self.lbsn_records = []
        self.lbsn_relationships = []
        self.null_island = 0
        self.log = logging.getLogger("__main__")
        self.ignore_non_geotagged = ignore_non_geotagged
        self.min_geoaccuracy = min_geoaccuracy
        self.skipped_low_geoaccuracy = 0

    def get_skipped_geoaccuracy(self):
        """Get count of records skipped due to low geoaccuracy"""
        return self.skipped_low_geoaccuracy

    def parse_json_record(self, json_string_dict, *_):
        """Parse a standard Instagram json record"""
        # clear any records from previous run
        self.lbsn_records.clear()
        if json_string_dict is None:
            return
        json_data = json_string_dict.get("graphql")
        if json_data:
            place_json = json_data.get("location")
        else:
            # Assume Place without header
            place_json = json_string_dict
        if place_json is None:
            return self.lbsn_records
        place_record, country_ref_pkey, city_ref_pkey = self.extract_place(place_json)
        # postRecord.post_geoaccuracy = twitterPostAttributes.geoaccuracy
        self.lbsn_records.append(place_record)
        # extract posts
        post_node_list = place_json.get("edge_location_to_media").get("edges")
        if post_node_list:
            self.extract_posts(post_node_list, place_record)
        # extract top posts
        top_post_node_list = place_json.get("edge_location_to_top_posts").get("edges")
        if top_post_node_list:
            self.extract_posts(
                top_post_node_list, place_record, country_ref_pkey, city_ref_pkey
            )
        # finally, return list of all extracted records
        return self.lbsn_records

    def extract_posts(
        self,
        node_list,
        place_record: lbsn.Place = None,
        country_ref_pkey=None,
        city_ref_pkey=None,
    ):
        for post_node in node_list:
            json_string_dict = post_node.get("node")
            if json_string_dict:
                self.extract_post(
                    json_string_dict, place_record, country_ref_pkey, city_ref_pkey
                )

    def extract_user(self, json_string_dict):
        user = json_string_dict
        user_record = HF.new_lbsn_record_with_id(
            lbsn.User(), user.get("id"), self.origin
        )
        return user_record

    def extract_post(
        self,
        json_string_dict: Dict[str, Any],
        place_record: lbsn.Place = None,
        country_ref_pkey=None,
        city_ref_pkey=None,
    ):
        post_guid = json_string_dict.get("id")
        if not HF.check_notice_empty_post_guid(post_guid):
            return None
        post_record = HF.new_lbsn_record_with_id(lbsn.Post(), post_guid, self.origin)
        user_record = None
        user_info = json_string_dict.get("owner")
        if user_info:
            # Get Post/Reaction Details of User
            user_record = self.extract_user(user_info)
        if user_record:
            self.lbsn_records.append(user_record)
        else:
            self.log.warning(
                f"No User record found for post: {post_guid} "
                f"(post saved without userid).."
            )

        # Check from upstream to update post attrs
        if place_record:
            # assign place accuracy, by default
            post_record.post_geoaccuracy = lbsn.Post.PLACE
            post_record.place_pkey.CopyFrom(place_record.pkey)
            post_record.post_latlng = place_record.geom_center
        else:
            post_record.post_geoaccuracy = None
        # referenced country and city
        if city_ref_pkey:
            post_record.city_pkey.CopyFrom(city_ref_pkey)
        if country_ref_pkey:
            post_record.country_pkey.CopyFrom(country_ref_pkey)
        # if still no geoinformation, send post to Null-Island
        if not post_record.post_latlng:
            if self.ignore_non_geotagged is True:
                return None
            else:
                self.null_island += 1
                post_record.post_latlng = "POINT(%s %s)" % (0, 0)
        if self.min_geoaccuracy:
            if not HF.geoacc_within_threshold(
                post_record.post_geoaccuracy, self.min_geoaccuracy
            ):
                self.skipped_low_geoaccuracy += 1
                return
        post_record.post_publish_date.CopyFrom(
            HF.json_date_timestamp_to_proto(json_string_dict.get("taken_at_timestamp"))
        )
        if user_record:
            post_record.user_pkey.CopyFrom(user_record.pkey)

        def value_count(x):
            return 0 if x is None else x

        post_record.post_comment_count = value_count(
            json_string_dict.get("edge_media_to_comment").get("count")
        )
        post_record.post_like_count = value_count(
            json_string_dict.get("edge_liked_by").get("count")
        )
        post_shortcode = json_string_dict.get("shortcode")
        post_record.post_url = f"http://www.instagram.com/p/{post_shortcode}"
        if json_string_dict.get("thumbnail_src"):
            post_record.post_thumbnail_url = json_string_dict.get("thumbnail_src")
        post_caption_edge = json_string_dict.get("edge_media_to_caption")
        if post_caption_edge:
            post_caption_edge_edges = post_caption_edge.get("edges")
            if post_caption_edge_edges and not len(post_caption_edge_edges) == 0:
                post_caption = post_caption_edge["edges"][0]["node"]["text"]
                post_record.post_body = post_caption.replace("\n", " ").replace(
                    "\r", ""
                )
                hashtags = HF.extract_hashtags_from_string(post_caption)
                if hashtags:
                    for hashtag in hashtags:
                        post_record.hashtags.append(hashtag)
        is_video = json_string_dict.get("is_video")
        if is_video:
            post_record.post_type = lbsn.Post.VIDEO
            post_record.post_views_count = value_count(
                json_string_dict.get("video_view_count")
            )
        else:
            post_record.post_type = lbsn.Post.IMAGE
        post_record.emoji.extend(HF.extract_emoji(post_record.post_body))
        self.lbsn_records.append(post_record)

    def extract_place(self, postplace_json):
        place = postplace_json
        place_id = place.get("id")
        if not place_id:
            self.log.warning(f"No PlaceGuid\n\n{place}")
            input("Press Enter to continue... (entry will be skipped)")
            return
        lon_center = place.get("lng")
        lat_center = place.get("lat")
        if lon_center is None or lat_center is None:
            # assign place to Null Island
            lon_center = 0
            lat_center = 0
            self.log.info(f"Empty lat or lng: Location {place_id}")
        # place_guid
        # For POIs, City is not available on Twitter
        place_record = HF.new_lbsn_record_with_id(lbsn.Place(), place_id, self.origin)
        place_record.geom_center = "POINT(%s %s)" % (lon_center, lat_center)
        place_name = place.get("name").replace("\n\r", "")
        # for some reason, twitter place entities sometimes contain
        # linebreaks or whitespaces. We don't want this.
        place_name = place.get("name").replace("\n\r", "")
        # remove multiple whitespace
        place_name = re.sub(" +", " ", place_name)
        place_record.name = place_name
        place_slug = place.get("slug")
        if place_slug:
            place_record.url = (
                f"https://www.instagram.com/explore/locations/"
                f"{place_id}/{place_slug}"
            )
        place_record.place_website = place.get("website")
        place_record.place_description = place.get("blurb")
        place_record.place_phone = place.get("phone")
        address_string = place.get("address_json")
        if address_string:
            address_json = json.loads(address_string)
            place_adr_street = address_json.get("street_address")
            place_zip_code = address_json.get("zip_code")
            place_city_name = address_json.get("city_name")
            place_region_name = address_json.get("region_name")
            place_country_name = address_json.get("US")
            place_record.address = ", ".join(
                filter(
                    None,
                    (
                        place_adr_street,
                        place_city_name,
                        place_region_name,
                        place_country_name,
                    ),
                )
            )
            place_record.zip_code = place_zip_code
        # extract directory
        directory_json = place.get("directory")
        if directory_json:
            country_json = directory_json.get("country")
            if country_json:
                cid = country_json.get("id")
                slug = country_json.get("slug")
                ref_country_record = HF.new_lbsn_record_with_id(
                    lbsn.Country(), cid, self.origin
                )
                ref_country_record.name = country_json.get("name")
                ref_country_record.url = (
                    f"https://www.instagram.com/explore/locations/{cid}/{slug}"
                )
                self.lbsn_records.append(ref_country_record)
            city_json = directory_json.get("city")
            if city_json:
                cid = city_json.get("id")
                slug = city_json.get("slug")
                ref_city_record = HF.new_lbsn_record_with_id(
                    lbsn.City(), cid, self.origin
                )
                ref_city_record.name = city_json.get("name")
                ref_city_record.url = (
                    f"https://www.instagram.com/explore/locations/{cid}/{slug}"
                )
                ref_city_record.country_pkey.CopyFrom(ref_country_record.pkey)
                self.lbsn_records.append(ref_city_record)
                place_record.city_pkey.CopyFrom(ref_city_record.pkey)
        country_ref_pkey = None
        if directory_json and country_json:
            country_ref_pkey = ref_country_record.pkey
        city_ref_pkey = None
        if directory_json and city_json:
            city_ref_pkey = ref_city_record.pkey
        return place_record, country_ref_pkey, city_ref_pkey



"""
Run Integration test based on yfcc data


This will check complete process of main() as if
invoked with

lbsntransform --origin 21 \
    --file_input \
    --input_path_url "https://example_url.com/yfcc100m_dataset.csv;https://example_url.com/yfcc100m_places.csv" \
    --dbpassword_output "eXamPlePassWoRd" \
    --dbuser_output "postgres" \
    --dbserveraddress_output "127.0.0.1:10432" \
    --dbname_output "rawdb" \
    --csv_delimiter $'\t' \
    --file_type "csv" \
    --zip_records \
    --mappings_path "./resources/mappings/" \
    --transferlimit 10000

Requirements:
- a local postgres container with lbsn rawdb must be available
https://gitlab.vgiscience.de/lbsn/databases/rawdb
- the database must be named rawdb_test
"""

import logging
import os
import sys
import traceback
from pathlib import Path

from lbsntransform.__main__ import main as lt_main  # type: ignore
from lbsntransform.config.config import BaseConfig  # type: ignore
from lbsntransform.input.load_data import LoadData  # type: ignore

# disable requests/urllib3 logging test urls in CI
logging.getLogger("urllib3").setLevel(logging.WARNING)


def system_integration_yfcc():
    """Test complete method integration using main()"""
    data_source_url: str = os.getenv("YFCC_URL")
    db_pass: str = os.getenv("RAWDB_TEST_PW")
    db_port: str = os.getenv("RAWDB_PORT")
    dbserveraddress_output: str = os.getenv("DB_IP")
    if not dbserveraddress_output:
        dbserveraddress_output = "127.0.0.1"
    # override default values with test case
    tm_cfg = BaseConfig()
    tm_cfg.origin = 21  # 2= Flickr, 1= YFCC
    tm_cfg.is_local_input = True
    tm_cfg.local_file_type = "csv"
    tm_cfg.input_path = data_source_url.split(";")
    tm_cfg.source_web = True
    tm_cfg.dbpassword_output = db_pass
    tm_cfg.dbname_output = "rawdb_test"
    tm_cfg.dbuser_output = "postgres"
    tm_cfg.dbserveraddress_output = dbserveraddress_output
    tm_cfg.dbserverport_output = db_port
    tm_cfg.csv_delim = "\t"
    tm_cfg.zip_records = True
    tm_cfg.mappings_path = Path.cwd() / "resources" / "mappings/"
    tm_cfg.transferlimit = 10000

    try:
        lt_main(tm_cfg)
        # sys.exit(0)
    except Exception:
        traceback.print_exc()
        sys.exit(1)

    # connect to test db and check results
    dbconn, dbcursor = LoadData.initialize_connection(
        tm_cfg.dbuser_output,
        tm_cfg.dbserveraddress_output,
        tm_cfg.dbname_output,
        tm_cfg.dbpassword_output,
        tm_cfg.dbserverport_output,
    )
    dbcursor.execute(
        """
        SELECT 'topical.post' AS table_name, count(*) FROM topical.post
        UNION
        SELECT 'spatial.country' AS table_name, count(*) FROM spatial.country
        UNION
        SELECT 'spatial.city' AS table_name, count(*) FROM spatial.city
        UNION
        SELECT 'spatial.place' AS table_name, count(*) FROM spatial.place
        UNION
        SELECT 'social.user' AS table_name, count(*) FROM social.user
        """
    )
    results = dict(dbcursor.fetchall())

    # compare with expected results for YFCC first 10000 records
    assert results["spatial.city"] == 1265
    assert results["topical.post"] == 3332
    assert results["spatial.country"] == 245
    assert results["social.user"] == 2991

    # cleanup step
    dbcursor.execute("""TRUNCATE social.origin CASCADE""")
    dbconn.commit()
    dbcursor.close()


if __name__ == "__main__":
    system_integration_yfcc()



# -*- coding: utf-8 -*-

"""Hll bases for spatial facet
"""

from typing import Union

import lbsnstructure as lbsn
from lbsntransform.output.hll import hll_bases as hll
from lbsntransform.tools.helper_functions import HelperFunctions as HF

from lbsntransform.output.hll.hll_functions import HLLFunctions as HLF

FACET = "spatial"


class LatLngBase(hll.HllBase):
    """Extends Base Class"""

    # base reference, eg.
    # e.g.: facet: temporal, base: date
    NAME = hll.HllBaseRef(facet=FACET, base="latlng")

    def __init__(self, record: lbsn.Post = None):
        super().__init__()
        # init key and any additional attributes
        self.key["latitude"] = None
        self.key["longitude"] = None
        self.attrs["latlng_geom"] = None
        # init additional metrics
        # beyond those defined inHllBase
        self.metrics["pud_hll"] = set()
        self.metrics["utl_hll"] = set()
        if record is None:
            # init empty
            return
        if isinstance(record, lbsn.Post):
            coordinates_geom = record.post_latlng
            coordinates = HF.get_coordinates_from_ewkt(coordinates_geom)
            self.key["latitude"] = coordinates.lat
            self.key["longitude"] = coordinates.lng
            # additional (optional) attributes
            # formatted ready for sql upsert
            self.attrs["latlng_geom"] = HF.return_ewkb_from_geotext(coordinates_geom)
        else:
            raise ValueError("Parsing of LatLngBase only supported " "from lbsn.Post")


class PlaceBase(hll.HllBase):
    """Extends HllBase Class"""

    NAME = hll.HllBaseRef(facet=FACET, base="place")

    def __init__(self, record: Union[lbsn.Post, lbsn.Place] = None):
        super().__init__()
        self.key["place_guid"] = None
        self.attrs["geom_center"] = None
        self.attrs["geom_area"] = None
        self.attrs["name"] = None
        self.metrics["pud_hll"] = set()
        self.metrics["utl_hll"] = set()
        if record is None:
            return
        if isinstance(record, lbsn.Post):
            # Post can be of Geoaccuracy "Place" without any
            # actual place id assigned (e.g. Flickr Geoaccuracy level < 10)
            # in this case, concat lat:lng as primary key
            coordinates_geom = record.post_latlng
            if not record.place_pkey.id:
                coordinates = HF.get_coordinates_from_ewkt(coordinates_geom)
                self.key["place_guid"] = HLF.hll_concat(
                    [coordinates.lat, coordinates.lng]
                )
            else:
                self.key["place_guid"] = record.place_pkey.id
            # additional (optional) attributes
            # formatted ready for sql upsert
            self.attrs["geom_center"] = HF.return_ewkb_from_geotext(coordinates_geom)
            # geom_area not available from lbsn.Post
        elif isinstance(record, lbsn.Place):
            coordinates_geom = record.geom_center
            coordinates = HF.get_coordinates_from_ewkt(coordinates_geom)
            self.key["place_guid"] = record.pkey.id
            # self.key['place_guid'] = HLF.hll_concat(
            #     [coordinates.lat, coordinates.lng])
            self.attrs["geom_center"] = HF.return_ewkb_from_geotext(coordinates_geom)
            self.attrs["geom_area"] = HF.return_ewkb_from_geotext(
                HF.null_check(record.geom_area)
            )
            self.attrs["name"] = record.name


class SpatialBase(hll.HllBase):
    """Intermediate spatial base class that extends HllBase
    for Place, City, Region and Country
    sharing common attributes
    """

    def __init__(
        self, record: Union[lbsn.Post, lbsn.Place, lbsn.City, lbsn.Country] = None
    ):
        super().__init__()
        self.key["guid"] = None
        self.attrs["name"] = None
        self.attrs["geom_center"] = None
        self.attrs["geom_area"] = None
        self.metrics["pud_hll"] = set()
        self.metrics["utl_hll"] = set()
        self.metrics["latlng_hll"] = set()
        if record is None:
            # init empty
            return
        name = None
        geom_area = None
        if isinstance(record, lbsn.Post):
            coordinates_geom = record.post_latlng
            coordinates = HF.get_coordinates_from_ewkt(coordinates_geom)
            # use concat lat:lng as key of no place_key available
            # this should later implement assignemnt based on area
            # intersection
            self.key["guid"] = HLF.hll_concat([coordinates.lat, coordinates.lng])
        elif isinstance(record, (lbsn.Place, lbsn.City, lbsn.Country)):
            name = HF.null_check(record.name)
            coordinates_geom = record.geom_center
            geom_area = record.geom_area
            # use key from place, city or country record
            self.key["guid"] = HLF.hll_concat_origin_guid(record)

        self.attrs["name"] = name
        self.attrs["geom_center"] = HF.return_ewkb_from_geotext(coordinates_geom)
        self.attrs["geom_area"] = HF.return_ewkb_from_geotext(geom_area)


class CityBase(SpatialBase):
    """Extends Spatial Base Class"""

    NAME = hll.HllBaseRef(facet=FACET, base="city")

    def __init__(self, record: lbsn.Post = None):
        super().__init__(record)
        # rename key accordingly
        self.key["city_guid"] = self.key.pop("guid")


class RegionBase(SpatialBase):
    """Extends Spatial Base Class"""

    NAME = hll.HllBaseRef(facet=FACET, base="region")

    def __init__(self, record: lbsn.Post = None):
        super().__init__(record)
        # rename key accordingly
        self.key["region_guid"] = self.key.pop("guid")


class CountryBase(SpatialBase):
    """Extends Spatial Base Class"""

    NAME = hll.HllBaseRef(facet=FACET, base="country")

    def __init__(self, record: lbsn.Post = None):
        super().__init__(record)
        # rename key accordingly
        self.key["country_guid"] = self.key.pop("guid")



# -*- coding: utf-8 -*-

"""
Config module for parsing input args for lbsntransform package.
"""

# pylint: disable=no-member

import argparse
import logging
from pathlib import Path
from typing import List, Tuple

from shapely import geos

import lbsnstructure as lbsn
from lbsntransform import __version__


class BaseConfig:
    """Base config class for handling typical input params"""

    def __init__(self):
        """Set Default Config options here

        - or define options as input args
        """
        self.origin = 0
        self.is_local_input = False
        self.local_file_type = 'json'
        self.input_path = None
        self.is_stacked_json = False
        self.is_line_separated_json = False
        self.dbuser_input = 'example-user-name'
        self.dbpassword_input = 'example-user-password'
        self.dbserveraddress_input = '222.22.222.22'
        self.dbserverport_input = 5432
        self.dbformat_input = 'json'
        self.dbname_input = 'test_db2'
        self.dbname_hllworker = None
        self.dbuser_hllworker = None
        self.dbpassword_hllworker = None
        self.dbserveraddress_hllworker = None
        self.dbserverport_hllworker = 5432
        self.dbname_hllworker = None
        self.dbuser_output = None
        self.dbformat_output = 'lbsn'
        self.dbpassword_output = None
        self.dbserveraddress_output = None
        self.dbserverport_output = 5432
        self.dbname_output = None
        self.transferlimit = None
        self.transfer_count = 50000
        self.number_of_records_to_fetch = 10000
        self.transfer_reactions = True
        self.disable_reactionpost_ref = False
        self.ignore_non_geotagged = False
        self.startwith_db_rownumber = None
        self.endwith_db_rownumber = None
        self.debug_mode = 'INFO'
        self.geocode_locations = False
        self.ignore_input_source_list = False
        self.input_lbsn_type = None
        self.map_relations = False
        self.csv_output = False
        self.csv_suppress_linebreaks = True
        self.csv_delim = None
        self.use_csv_dictreader = None
        self.recursive_load = False
        self.skip_until_file = None
        self.min_geoaccuracy = None
        self.logging_level = logging.INFO
        self.source_web = False
        self.skip_until_record = None
        self.zip_records = None
        self.include_lbsn_objects = []
        self.include_lbsn_bases = None
        self.override_lbsn_query_schema = None
        self.mappings_path = None
        self.dry_run = None
        self.hmac_key = None
        self.commit_volume = None

        BaseConfig.set_options()

    @staticmethod
    def get_arg_parser(
        parser: argparse.ArgumentParser = None) -> argparse.ArgumentParser:
        """Define lbsntransform cli arguments

        Arguments:
            parser: Optional parser with extended arguments. This
            can be used to include lbsntransform in another package.

        Note: All args are optional, but some groups need to be defined
            together.

        Important: These settings are used to generate the command line docs [1].
        [1]: https://lbsn.vgiscience.org/lbsntransform/docs/argparse/args/
        """
        if parser is None:
            parser = argparse.ArgumentParser()
        parser.add_argument(
            '--version',
            action='version',
            version=f'{__version__}')
        parser.add_argument(
            '-o', "--origin",
            default=0,
            help='Input source type (id). '
            '  '
            '  '
            '* Defaults to `0`: LBSN  '
            '  '
            'Other possible values:  '
            '* `1` - Instagram  '
            '* `2` - Flickr  '
            '* `21` - Flickr YFCC100M  '
            '* `3` - Twitter  ',
            type=int)
        parser.add_argument(
            "--dry-run",
            action='store_true',
            help='Perform a trial run '
            '  '
            'with no changes made '
            'to database/output')
        # Local Input
        local_input_args = parser.add_argument_group('Local Input')
        local_input_args.add_argument(
            '-l', "--file_input",
            action='store_true',
            help='This flag enables file input  '
            '  '
            '(instead of reading data from a database). '
            '  '
            '  '
            '* To specify which files to process, see '
            'parameter `--input_path_url`.  '
            '* To specify file types, e.g. whether to '
            'process data from `json` '
            'or `csv`, or from URLs,  '
            '  see `--file_type`  ')
        local_input_args.add_argument(
            "--file_type",
            default='json',
            help='Specify filetype  '
            '  '
            ' (`json`, `csv` etc.) '
            '  '
            '  '
            '* only applies if `--file_input` is used.  ',
            type=str)
        local_input_args.add_argument(
            "--input_path_url",
            default="01_Input",
            help='Path to input folder.  '
            '  '
            '* If not provided, subfolder `./01_Input/` '
            ' will be used.  '
            '* You can also provide a web-url, '
            'starting with `http(s)`  '
            '* URLs will be accessed using '
            '`requests.get(url, stream=True)`.  '
            '* To separate multiple urls, use '
            'semicolon (`;`). In this case, see also '
            '`--zip_records`.  ',
            type=str)
        local_input_args.add_argument(
            "--is_stacked_json",
            action='store_true',
            default=False,
            help='Input is stacked json. '
            '  '
            '  '
            '* The typical form of json is '
            '`[{json1},{json2}]`  '
            '* If `--is_stacked_json` is set, '
            'it will process stacked jsons in the form '
            'of `{json1}{json2}` (no comma)  ')
        local_input_args.add_argument(
            "--is_line_separated_json",
            action='store_true',
            default=False,
            help='Json is line separated '
            '  '
            '  '
            '* The typical form is '
            '`[{json1},{json2}]`  '
            '* If `--is_line_separated_json` is set, '
            'it will process stacked jsons in the form '
            'of `{json1}\n{json2}` (with linebreak)  '
            '* Unix style linebreaks (CR) will be used '
            'across platforms  '
            '* Windows users, use (e.g.) notepad++ to '
            'convert from Windows style linebreaks (CRLF)  ')
        # HLL Worker
        hllworker_args = parser.add_argument_group('HLL Worker')
        hllworker_args.add_argument(
            "--dbpassword_hllworker",
            help='Password for hllworker db  '
            '  '
            '  '
            '* If reading data into `hlldb`, all '
            'HLL Worker parameters must be supplied by'
            'default.  '
            '* You can substitute hlldb parameters here  '
            '* In this case, lbsntransform will use '
            'hlldb to convert and union hll sets '
            '_and_ to store output results  '
            '* Currently, this re-use of hlldb requires '
            'to supply the same set of parameters twice  '
            '* For separation of concerns, it is recommended '
            'to use a separate HLL Worker database  ',
            type=str)
        hllworker_args.add_argument(
            "--dbuser_hllworker",
            default="postgres",
            help='Username for hllworker db.',
            type=str)
        hllworker_args.add_argument(
            "--dbserveraddress_hllworker",
            help='IP for hllworker db  '
            '  '
            '  '
            '* e.g. `111.11.11.11`  '
            '* Optionally add '
            'port the to use, e.g. `111.11.11.11:5432`.  '
            '* `5432` is the default port  ',
            type=str)
        hllworker_args.add_argument(
            "--dbname_hllworker",
            help='DB name for hllworker db '
            '  '
            '  '
            '* e.g. `hllworkerdb`  ',
            type=str)
        # DB Output
        dboutput_args = parser.add_argument_group(
            'DB Output')
        dboutput_args.add_argument(
            '-p', "--dbpassword_output",
            help='Password for out-db  '
            '  '
            '(postgres raw/hll db)',
            type=str)
        dboutput_args.add_argument(
            '-u', "--dbuser_output",
            default="postgres",
            help='Username for out-db.  '
            '  '
            'Default: `example-user-name2`',
            type=str)
        dboutput_args.add_argument(
            '-a', "--dbserveraddress_output",
            help='IP for output db, '
            '  '
            '  '
            '* e.g. `111.11.11.11`  '
            '* Optionally add '
            'port to use, e.g. `111.11.11.11:5432`.  '
            '* `5432` is the default port  ',
            type=str)
        dboutput_args.add_argument(
            '-n', "--dbname_output",
            help='DB name for output db '
            '  '
            '  '
            '* e.g. `rawdb` or `hlldb`  ',
            type=str)
        dboutput_args.add_argument(
            "--dbformat_output",
            default="lbsn",
            help='Format of the out-db. '
            '  '
            '  '
            '* Either `hll` or `lbsn`.  '
            '* This setting '
            'affects how data is stored, either '
            'in anonymized and aggregate form (`hll`), or in '
            'the [lbsn raw structure](https://lbsn.vgiscience.org/structure/) '
            '(`lbsn`).  ',
            type=str)
        # DB Input
        dbinput_args = parser.add_argument_group(
            'DB Input')
        dbinput_args.add_argument(
            "--dbpassword_input",
            help='Password for input-db',
            type=str)
        dbinput_args.add_argument(
            "--dbuser_input",
            default="postgres",
            help='Username for input-db.',
            type=str)
        dbinput_args.add_argument(
            "--dbserveraddress_input",
            help='IP for input-db, '
            '  '
            '  '
            '* e.g. `111.11.11.11`  '
            '* Optionally add port to use, e.g. '
            '`111.11.11.11:5432`.  '
            '* `5432` is the default port  ',
            type=str)
        dbinput_args.add_argument(
            "--dbname_input",
            help='DB name for input-db, '
            '  '
            '  '
            '* e.g.: `rawdb`  ',
            type=str)
        dbinput_args.add_argument(
            "--dbformat_input",
            default="json",
            help='Format of the input-db. '
            '  '
            '  '
            '* Either `lbsn` or `json`  '
            '* If lbsn is used, the native lbsn raw '
            'input mapping (`0`) will be used  '
            '* If `json` is used, a custom mapping for '
            'json must be provided, for mapping database '
            'json\'s to the lbsn structure. See '
            '[input mappings](https://lbsn.vgiscience.org/lbsntransform/docs/mappings/#input-mappings)  ',
            type=str)
        # Additional args
        settings_args = parser.add_argument_group(
            'Additional settings')
        settings_args.add_argument(
            '-t', "--transferlimit",
            help='Abort after x records. '
            '  '
            '  '
            '* This can be used to limit the number of '
            'records that will be processed.  '
            '* e.g. `--transferlimit 10000` will process '
            'the first 10000 input records  '
            '* Defaults to None (= process all)  '
            '* Note that one input record can map to '
            'many output records. This number applies to '
            'the number of input records, not the output count.  ',
            type=int)
        settings_args.add_argument(
            "--transfer_count",
            default=50000,
            help='Transfer batch limit x. '
            '  '
            '  '
            '* Defines after how many '
            'parsed records the results will be '
            'transferred to the DB.  '
            '* Defaults to 50000  '
            '* If you have a slow server, but a fast machine, larger values '
            'improve speed because duplicate '
            'check happens in Python, and not in '
            'Postgres coalesce;  '
            '* However, larger values require '
            'more local memory. If you have a fast server, but a slow machine, '
            'try if a smaller batch `--transfer_count` (e.g. 5000) improves speed.  '
            '  '
            '  '
            '!!! note  '
            '    Use `--transferlimit` to limit the '
            '    total number of records transferred. `--transfer_count` '
            '    instead defines the _batch_ count that is used to transfer '
            '    data incrementally.  ',
            type=int)
        settings_args.add_argument(
            "--commit_volume",
            default=None,
            help='After x commit_volume, changes (transactions) will be written to '
            'the output database (a '
            '[Postgres COMMIT](https://www.postgresql.org/docs/current/sql-commit.html)). '
            '  '
            '  '
            'Note that updated entries in the output database are only written '
            'from the WAL buffer after a commit. '
            '  '
            '  '
            '* Default for rawdb: 10000  '
            '* Default for hlldb: 100000  '
            '  '
            '  '
            '!!! warning  '
            '    If you have concurrent writes to the DB (e.g. multiple '
            '    lbsntransform processes) and if you see transaction deadlocks, '
            '    reduce the commit_volume.  ',
            type=int)
        settings_args.add_argument(
            "--records_tofetch",
            default=10000,
            help='Fetch x records /batch. '
            '  '
            '  '
            '* If retrieving data from a db (`lbsn`), '
            'limit the number of records to fetch at once.  '
            '* Defaults to 10000  ',
            type=int)
        settings_args.add_argument(
            "--disable_transfer_reactions",
            action='store_true',
            help='Disable reactions. '
            '  '
            '  '
            '* If set, processing of lbsn reactions will be skipped,  '
            '* only original posts are transferred.  '
            '* This is usefull to reduce '
            'processing and data footprint for some service data, '
            'e.g. for Twitter, with a large '
            'number of reactions containing little original content.  ')
        settings_args.add_argument(
            "--disable_reaction_post_referencing",
            action='store_true',
            default=False,
            help='Disable reactions-refs. '
            '  '
            '  '
            'Enable this option in args '
            'to prevent empty posts being stored '
            'due to Foreign-Key-Exists Requirement. '
            '  '
            'Possible parameters:  '
            '* `0` = Save Original Tweets of Retweets as `posts`;  '
            '* `1` = do not store Original Tweets of Retweets;  '
            '* `2` = !Not implemented: Store Original Tweets of Retweets as '
            '`post_reactions`  ')
        settings_args.add_argument(
            "--ignore_non_geotagged",
            action='store_true',
            help='Ignore none-geotagged. '
            '  '
            '  '
            'If set, posts that are not geotagged '
            'are ignored during processing.')
        settings_args.add_argument(
            "--startwith_db_rownumber",
            help='Start with db row x. '
            '  '
            '  '
            'If transferring from a databse (input), this flag can be used to '
            'resume processing (e.g.) '
            'if a transfer has been aborted. '
            '  '
            '  '
            '* Provide a number (row-id) to start '
            'processing from live db.  '
            '* If input db type '
            'is `lbsn`, this is the primary key, without the `origin_id`, '
            '(e.g. the `post_guid`, `place_guid` etc.).  '
            '* This flag will only work if processing a '
            'single lbsn object (e.g. `--include_lbsn_objects "post"`).  '
            '  '
            'Example:  '
            '`--startwith_db_rownumber "123456789"`  '
            'will lead to the first batch-query from the DB looking like this:  '
            '```sql'
            'SELECT * FROM topical."post"  '
            'WHERE post_guid > \'123456789\'  '
            'ORDER BY post_guid ASC  '
            'LIMIT 10000;  '
            '```'
            '  '
            '  ',
            type=str)
        settings_args.add_argument(
            "--endwith_db_rownumber",
            help='End with db row x. '
            '  '
            '  '
            'Provide a number (row-id) to end '
            'processing from live db', type=int)
        settings_args.add_argument(
            "--debug_mode",
            help='Enable debug mode.',
            type=str)
        settings_args.add_argument(
            "--geocode_locations",
            help='Path to loc-geocodes. '
            '  '
            '  '
            '* Provide path to a CSV file with '
            'location geocodes  '
            '* CSV Header must be: '
            '`lat, lng, name`).  '
            '* This can be used in mappings to assign '
            'coordinates (lat, lng) '
            'to use provided locations as text  ',
            type=str)
        settings_args.add_argument(
            "--ignore_input_source_list",
            help='Path to input ignore. '
            '  '
            '  '
            'Provide a path to a list of input_source '
            'types that will be ignored (e.g. to '
            'ignore certain bots etc.)',
            type=str)
        settings_args.add_argument(
            "--mappings_path",
            help='Path mappings folder. '
            '  '
            '  '
            'Provide a path to a custom folder '
            'that contains '
            'one or more input mapping modules (`*.py`). '
            '  '
            '  '
            '* Have a look at the two sample mappings '
            'in [the resources folder](https://gitlab.vgiscience.de/lbsn/lbsntransform/-/tree/master/resources/mappings).  '
            '* See how to define custom input mappings '
            'in the [docs](https://lbsn.vgiscience.org/lbsntransform/docs/mappings/#input-mappings)  ',
            type=str)
        settings_args.add_argument(
            "--input_lbsn_type",
            help='Input sub-type '
            '  '
            '  '
            '* e.g. `post`, `profile`, '
            '`friendslist`, `followerslist` etc.  '
            '* This can be used to select an appropiate '
            'mapping procedure in a single mapping module.  ',
            type=str)
        settings_args.add_argument(
            "--map_full_relations",
            action='store_true',
            help='Map full relations. '
            '  '
            '  '
            'Set to true to map full relations, '
            'e.g. many-to-many relationships, '
            'such as `user_follows`, '
            '`user_friend`, or `user_mentions` etc. are '
            'mapped in a separate table. '
            'Defaults to False.')
        settings_args.add_argument(
            "--csv_output",
            action='store_true',
            help='Store to local CSV. '
            '  '
            '  '
            'If set, will store all '
            'submit values to local CSV instead. '
            'Currently, this type of output is not available.')
        settings_args.add_argument(
            "--csv_allow_linebreaks",
            action='store_true',
            help=repr('Disable linebreak-rem. '
            '  '
            '  '
            'If set, will not '
            'remove intext-linebreaks (`\r` or `\n`) '
            'in output CSVs')[1:-1])
        settings_args.add_argument(
            "--csv_delimiter",
            default=",",
            help=repr(
                'CSV delimiter. '
                '  '
                '  '
                '* Provide the CSV delimiter to be used.  '
                '* Default is comma (`,`).  '
                '* Note: to pass tab, '
                'use variable substitution (`$"\t"`)  ')[1:-1],
            type=str)
        settings_args.add_argument(
            "--use_csv_dictreader",
            action='store_true',
            help='Use csv.DictReader. '
            '  '
            '  '
            'By default, CSVs will be read line by line,  '
            'using the standard csv.reader().  '
            '  '
            'This will enable '
            '[csv.DictReader()](https://docs.python.org/3/library/csv.html#csv.DictReader),  '
            'which allows to access CSV fields by name in mappings.  '
            '  '
            'A CSV with a header is required for this setting to work.  '
            '  '
            'Note that `csv.DictReader()` may be slower than the default '
            '`csv.reader()`.')
        settings_args.add_argument(
            "--recursive_load",
            action='store_true', default=False,
            help='Recursive local sub dirs. '
            '  '
            '  '
            'If set, process input directories '
            'recursively (default depth: `2`)')
        settings_args.add_argument(
            "--skip_until_file",
            help='Skip until file x. '
            '  '
            '  '
            'If local input, skip all files '
            'until file with name `x` appears '
            '(default: start immediately)',
            type=str)
        settings_args.add_argument(
            "--skip_until_record",
            help='Skip until record x. '
            '  '
            '  '
            'If local input, skip all records '
            'until record `x` '
            '(default: start with first)',
            type=int)
        settings_args.add_argument(
            "--zip_records",
            action='store_true', default=False,
            help='Zip records parallel. '
            '  '
            '  '
            '* Use this flag to zip records of '
            'multiple input files  '
            '* e.g. `List1[A,B,C]`, `List2[1,2,3]` will be '
            'combined (zipped) on read to '
            '`List[A1,B2,C3]`  ')
        settings_args.add_argument(
            "--min_geoaccuracy",
            help='Min geoaccuracy to use '
            '  '
            '  '
            'Set to `latlng`, `place`, '
            'or `city` to limit processing of records based '
            'on mininum geoaccuracy (default: no limit)',
            type=str)
        settings_args.add_argument(
            "--include_lbsn_objects",
            help='lbsn objects to process '
            '  '
            '  '
            'If processing from lbsn db (`rawdb`), '
            'provide a comma separated list '
            'of [lbsn objects][1] to include. '
            '  '
            'May contain:  '
            '* origin  '
            '* country  '
            '* city  '
            '* place  '
            '* user_groups  '
            '* user  '
            '* post  '
            '* post_reaction  '
            '  '
            'Notes:  '
            '* Excluded objects will not be queried, but empty '
            'objects may be created due to referenced '
            'foreign key relationships.  '
            '* Defaults to '
            '`origin,post`  '
            '[1]: https://lbsn.vgiscience.org/structure/',
            type=str)
        settings_args.add_argument(
            "--include_lbsn_bases",
            help='lbsn bases to update '
            '  '
            '  '
            'If the target output type is `hll`, '
            'provide a comma separated list '
            'of lbsn bases to include/update/store to. '
            '  '
            '  '
            'Currently supported:  '
            '* hashtag  '
            '* emoji  '
            '* term  '
            '* _hashtag_latlng  '
            '* _term_latlng  '
            '* _emoji_latlng  '
            '* _month_hashtag  '
            '* _month_hashtag_latlng  '
            '* _month_latlng  '
            '* monthofyear  '
            '* month  '
            '* dayofmonth  '
            '* dayofweek  '
            '* hourofday  '
            '* year  '
            '* month  '
            '* date  '
            '* timestamp  '
            '* country  '
            '* region  '
            '* city  '
            '* place  '
            '* latlng  '
            '* community  '
            '  '
            'Bases not included will be skipped. Per '
            'default, **no bases** will be considered. '
            '  '
            '  '
            'Example:  '
            '```bash'
            '--include_lbsn_bases hashtag,place,date,community  '
            '```'
            '  '
            'This will update entries in the Postgres '
            '[hlldb](https://gitlab.vgiscience.de/lbsn/databases/hlldb) tables '
            '`social.community`. `topical.hashtag`, `spatial.place`, '
            '`temporal.date` and non-existing entries will be created, existing '
            'ones will be updated (a `hll_union`). '
            '  '
            '  '
            'See the structure definition in SQL '
            '[here](https://gitlab.vgiscience.de/lbsn/structure/hlldb/-/blob/master/structure/98-create-tables.sql) '
            'for a full list of hlldb table structures.  '
            '  '
            'Argument only allowed one time.',
            type=str)
        settings_args.add_argument(
            "--override_lbsn_query_schema",
            help='Override schema and table name '
            '  '
            '  '
            'This can be used to redirect lbsn queries on '
            'the given object from input db to a specific schema/table '
            'such as a materialized view. '
            '  '
            '  '
            'This can be usefull (e.g.) to limit '
            'processing of input data to a specific '
            'query.  '
            '  '
            'Format is `lbsn_type,schema.table`.  '
            '  '
            'Example:  '
            '```bash'
            '--override_lbsn_query_schema post,mviews.mypostquery  '
            '```'
            'Argument can be used multiple times.',
            action='append',
            type=str)
        settings_args.add_argument(
            "--hmac_key",
            help='Override db hmac key '
            '  '
            '  '
            'The hmac key that is used for cryptographic hashing '
            'during creation of HLL sets. Override what is '
            'set in hllworker database here. '
            '  '
            '  '
            'Remember to re-use the same hmac key for any '
            'consecutive update of HLL sets. '
            '  '
            '  '
            'The crypt.salt variable can also be set (temporarily or permanently) in the '
            'hll worker database itself. '
            '  '
            'Example:  '
            '```sql'
            'ALTER DATABASE hllworkerdb SET crypt.salt = \'CRYPTSALT\';  '
            '```'
            '  '
            '  '
            'Further information is available in the [YFCC HLL tutorial][2]. '
            '  '
            '[2]: https://lbsn.vgiscience.org/tutorial/yfcc-geohash/#prepare-query-and-cryptographic-hashing',
            action='append',
            type=str)
        return parser

    def parse_args(self, args: List = None):
        """Process input *args

        Arguments:
            args: Optional prepopulated list of args. Can be used to define
            and extend arguments outside of lbsntransform
        """

        parser = BaseConfig.get_arg_parser()
        args = parser.parse_args(args)
        if args.file_input:
            self.is_local_input = True
            self.local_file_type = args.file_type
            if args.is_stacked_json:
                self.is_stacked_json = True
            if args.is_line_separated_json:
                self.is_line_separated_json = True
            if not args.input_path_url:
                self.input_path = Path.cwd() / "01_Input"
                print(f"Using Path: {self.input_path}")
            else:
                if str(args.input_path_url).startswith("http"):
                    self.input_path = args.input_path_url.split(";")
                    self.source_web = True
                else:
                    input_path = Path(args.input_path_url)
                    self.input_path = input_path
        else:
            self.dbuser_input = args.dbuser_input
            self.dbpassword_input = args.dbpassword_input
            input_ip, input_port = BaseConfig.get_ip_port(
                args.dbserveraddress_input)
            self.dbserveraddress_input = input_ip
            if input_port:
                self.dbserverport_input = input_port
            self.dbname_input = args.dbname_input
        if args.dry_run:
            self.dry_run = True
        if args.csv_output:
            raise NotImplementedError(
                "CSV output is currently not available.")
        if args.dbformat_input:
            self.dbformat_input = args.dbformat_input
        if args.origin:
            self.origin = args.origin
        if args.geocode_locations:
            self.geocode_locations = Path(
                args.geocode_locations)
        if args.ignore_input_source_list:
            self.ignore_input_source_list = Path(
                args.ignore_input_source_list)    
        if not self.csv_output and args.dbuser_output:
            self.dbuser_output = args.dbuser_output
            self.dbpassword_output = args.dbpassword_output
            output_ip, output_port = BaseConfig.get_ip_port(
                args.dbserveraddress_output)
            self.dbserveraddress_output = output_ip
            if output_port:
                self.dbserverport_output = output_port
            self.dbname_output = args.dbname_output
        if not self.csv_output and args.dbformat_output:
            self.dbformat_output = args.dbformat_output
            if self.dbformat_output == "hll":
                try:
                    self.dbname_hllworker = args.dbname_hllworker
                    self.dbuser_hllworker = args.dbuser_hllworker
                    self.dbpassword_hllworker = args.dbpassword_hllworker
                    hllworker_ip, hllworker_port = BaseConfig.get_ip_port(
                        args.dbserveraddress_hllworker)
                    self.dbserveraddress_hllworker = hllworker_ip
                    if hllworker_port:
                        self.dbserverport_hllworker = hllworker_port
                except ValueError:
                    raise ValueError(
                        "If '--dbformat_output hll' is used, also provide "
                        "a HLL Worker Connection (--dbname_hllworker, "
                        "--dbuser_hllworker, --dbpassword_hllworker, "
                        "--dbserveraddress_hllworker). You can also use the "
                        "output hlldb again here.")
        if args.transferlimit:
            self.transferlimit = args.transferlimit
            if self.transferlimit == 0:
                self.transferlimit = None
        if args.commit_volume:
            self.commit_volume = args.commit_volume
            if self.commit_volume == 0:
                self.commit_volume = None
        if args.transfer_count:
            self.transfer_count = args.transfer_count
        if args.records_tofetch:
            self.number_of_records_to_fetch = args.records_tofetch
        if args.disable_transfer_reactions:
            self.transfer_reactions = False
        if args.disable_reaction_post_referencing:
            self.disable_reactionpost_ref = True
        if args.ignore_non_geotagged:
            self.ignore_non_geotagged = True
        if args.startwith_db_rownumber:
            # hack: converts to int (row-count) if possible
            # otherwise uses str (row-id lookup)
            # TODO: separate into different cli args
            self.startwith_db_rownumber = self.int_or_str(
                args.startwith_db_rownumber)
        if args.endwith_db_rownumber:
            self.endwith_db_rownumber = args.endwith_db_rownumber
        if args.debug_mode:
            self.debug_mode = args.debug_mode
        if args.input_lbsn_type:
            self.input_lbsn_type = args.input_lbsn_type
        if args.map_full_relations:
            self.map_relations = True
        if args.csv_allow_linebreaks:
            self.csv_suppress_linebreaks = False
        if args.csv_delimiter:
            self.csv_delim = args.csv_delimiter
        if args.use_csv_dictreader:
            self.use_csv_dictreader = args.use_csv_dictreader
        if args.recursive_load:
            self.recursive_load = True
        if args.skip_until_file:
            self.skip_until_file = args.skip_until_file
        if args.zip_records:
            self.zip_records = True
        if args.skip_until_record:
            self.skip_until_record = args.skip_until_record
        if args.mappings_path:
            self.mappings_path = Path(args.mappings_path)
        if args.min_geoaccuracy:
            self.min_geoaccuracy = self.check_geoaccuracy_input(
                args.min_geoaccuracy)
        if args.include_lbsn_objects:
            self.include_lbsn_objects = \
                args.include_lbsn_objects.lower().split(",")
        else:
            self.include_lbsn_objects = ["origin,post"]
        if args.include_lbsn_bases:
            self.include_lbsn_bases = args.include_lbsn_bases.split(",")
        if args.override_lbsn_query_schema:
            self.override_lbsn_query_schema = self.compile_schema_override(
                args.override_lbsn_query_schema)
        if args.hmac_key:
            self.hmac_key = args.hmac_key

    @classmethod
    def compile_schema_override(
        cls, override_lbsn_query_schema: List[str]) -> List[Tuple[str, str]]:
        """Split lbsn_type from schema.table override"""
        lbsn_query_schema = []
        for override in override_lbsn_query_schema:
            try:
                lbsn_type, schema_table_override = override.split(",")
            except ValueError as e:
                raise ValueError(
                    f"Cannot split lbsn_type from schema.table ({override}). "
                    f"Make sure override_lbsn_query_schema entries are "
                    f"formatted correctly, e.g. lbsn_type,schema.table") from e
            lbsn_query_schema.append(
                (lbsn_type.lower(), schema_table_override.lower()))                         
        return lbsn_query_schema

    @staticmethod
    def check_geoaccuracy_input(geoaccuracy_string):
        """Checks geoaccuracy input string and matches
        against proto buf spec
        """
        if geoaccuracy_string == "latlng":
            return lbsn.Post.LATLNG
        if geoaccuracy_string == "place":
            return lbsn.Post.PLACE
        if geoaccuracy_string == "city":
            return lbsn.Post.CITY
        print("Unknown geoaccuracy.")
        return None

    @staticmethod
    def get_ip_port(ip_string: str):
        """Gets IP and port, if available

        Arguments:
            ip_string {str} -- String containing IPV4 and
            possibly port attached with : character
        """
        port = None
        ip_addr = None
        ip_string_spl = ip_string.split(":")
        if len(ip_string_spl) == 2:
            ip_addr = ip_string_spl[0]
            port = ip_string_spl[1]
        else:
            ip_addr = ip_string
        return ip_addr, port

    @staticmethod
    def set_options():
        """Includes global options in other packages to be set
        prior execution"""
        # tell shapely to include the srid when generating WKBs
        geos.WKBWriter.defaults["include_srid"] = True

    @classmethod
    def int_or_str(cls, value_str: str):
        """Returns int if value is of type int, otherwise str"""
        try:
            return int(value_str)
        except:
            return value_str



# -*- coding: utf-8 -*-

"""Hll bases for social facet
"""

from typing import Tuple, Union

import lbsnstructure as lbsn
from lbsntransform.output.hll import hll_bases as hll

from lbsntransform.output.hll.hll_functions import HLLFunctions as HLF

FACET = "social"


class SocialBase(hll.HllBase):
    """Intermediate social base class that extends HllBase
    for common social base attributes
    """

    def __init__(self):
        super().__init__()
        self.metrics["latlng_hll"] = set()
        self.metrics["place_hll"] = set()
        self.metrics["pud_hll"] = set()


class UserBase(SocialBase):
    """Extends Social Base Class"""

    NAME = hll.HllBaseRef(facet=FACET, base="user")

    def __init__(self, record: lbsn.User = None):
        super().__init__()
        self.key["user_guid"] = None
        if record is None:
            # init empty
            return
        # TODO: implement one-way-hashing
        self.key["user_guid"] = HLF.hll_concat_origin_guid(record)


class FriendsBase(SocialBase):
    """Extends Social Base Class"""

    NAME = hll.HllBaseRef(facet=FACET, base="friends")

    def __init__(self, friends_record: Tuple[lbsn.User, lbsn.User] = None):
        super().__init__()
        self.key["user_guid"] = None
        self.key["user_guid_friend"] = None
        if friends_record is None:
            # init empty
            return
        # TODO: implement one-way-hashing
        self.key["user_guid"] = HLF.hll_concat_origin_guid(friends_record[0])
        self.key["user_guid_friend"] = HLF.hll_concat_origin_guid(friends_record[1])


class CommunityBase(SocialBase):
    """Extends Social Base Class"""

    NAME = hll.HllBaseRef(facet=FACET, base="community")

    def __init__(self, record: Union[lbsn.Origin, lbsn.Post] = None):
        super().__init__()
        self.attrs["name"] = None
        self.key["id"] = None
        if record is None:
            # init empty
            return
        if hasattr(record, "origin_id"):
            self.key["id"] = record.origin_id
            self.attrs["name"] = record.DESCRIPTOR.name
        else:
            self.key["id"] = record.pkey.origin.origin_id


class CultureBase(SocialBase):
    """Extends Social Base Class"""

    NAME = hll.HllBaseRef(facet=FACET, base="culture")

    def __init__(self, record: lbsn.Language = None):
        super().__init__()
        self.key["id"] = None
        if record is None:
            # init empty
            return
        self.key["id"] = record.language_short



"""
Tests for command line interface (CLI).
"""
from importlib.metadata import version
from os import linesep

from .cli_test_helpers import shell


def test_version():
    """
    Does --version display information as expected?
    """
    expected_version = version("lbsntransform")
    result = shell("lbsntransform --version")

    assert result.stdout == f"{expected_version}{linesep}"
    assert result.exit_code == 0



"""
Tests for command line interface (CLI).
"""
import unittest

from lbsntransform.tools.helper_functions import HelperFunctions as HF  # type: ignore


class TestExtractors(unittest.TestCase):
    """Test various string operations from HelperFunctions"""

    def test_mentions(self):
        """
        Is the user mentioned extracted from the string?
        """
        test_body = "/u/Exz3ssion, vielen Dank f\u00fcr deinen Beitrag. "
        # test Reddit-style @-mentions
        expected_user = "Exz3ssion"
        result = HF.extract_user_mentions(test_body)
        assert expected_user in result
        # test Twitter-style @-mentions
        test_body = "@Exz3ssion, vielen Dank f\u00fcr deinen Beitrag."
        result = HF.extract_atmentions_from_string(test_body)
        assert expected_user in result
        # test Reddit user with a Minus sign
        test_body = (
            "/u/goto-reddit was genau ist eigentlich dein Problem? /u/goto_reddit"
        )
        result = HF.extract_user_mentions(test_body)
        expected_user = ["goto-reddit", "goto_reddit"]
        for user in expected_user:
            assert user in result

    def test_hashtags(self):
        """
        Are all of the hastags extracted from the string?
        """
        test_body = "#germany🇩🇪 #life #green #1234"
        result = HF.extract_hashtags_from_string(test_body)
        expected_tags = {"germany", "life", "green", "1234"}

        unittest.TestCase.assertSetEqual(self, result, expected_tags)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

"""Hll bases for temporal facet
"""

import lbsnstructure as lbsn
from lbsntransform.output.hll import hll_bases as hll

from lbsntransform.output.hll.hll_functions import HLLFunctions as HLF
from lbsntransform.tools.helper_functions import HelperFunctions as HF

FACET = "temporal"


class TimestampBase(hll.HllBase):
    """Extends Base Class"""

    NAME = hll.HllBaseRef(facet=FACET, base="timestamp")

    def __init__(self, record: lbsn.Post = None):
        super().__init__()
        self.key["timestamp"] = None
        if record is None:
            # init empty
            return
        post_date_time = HLF.merge_dates_post(record)
        if post_date_time:
            self.key["timestamp"] = post_date_time


class DateBase(hll.HllBase):
    """Extends Base Class"""

    NAME = hll.HllBaseRef(facet=FACET, base="date")

    def __init__(self, record: lbsn.Post = None):
        super().__init__()
        self.key["date"] = None
        self.attrs["name"] = None
        if record is None:
            # init empty
            return
        post_date_time = HLF.merge_dates_post(record)
        if post_date_time:
            # optional: add name of date here (e.g. "New Year's Eve")
            date = post_date_time.date()
            self.key["date"] = date


class MonthBase(hll.HllBase):
    """Extends Temporal Base Class"""

    NAME = hll.HllBaseRef(facet=FACET, base="month")

    def __init__(self, record: lbsn.Post = None):
        super().__init__()
        self.key["year"] = None
        self.key["month"] = None
        if record is None:
            # init empty
            return
        post_date_time = HLF.merge_dates_post(record)
        if post_date_time:
            date = post_date_time.date()
            self.key["year"] = date.year
            self.key["month"] = date.month


class YearBase(hll.HllBase):
    """Extends Base Class"""

    NAME = hll.HllBaseRef(facet=FACET, base="year")

    def __init__(self, record: lbsn.Post = None):
        super().__init__()
        self.key["year"] = None
        if record is None:
            # init empty
            return
        post_date_time = HLF.merge_dates_post(record)
        if post_date_time:
            date = post_date_time.date()
            self.key["year"] = date.year


class TimeofdayBase(hll.HllBase):
    """Extends Base Class"""

    NAME = hll.HllBaseRef(facet=FACET, base="timeofday")

    def __init__(self, record: lbsn.Post = None):
        super().__init__()
        self.key["timeofday"] = None
        if record is None:
            # init empty
            return
        post_date_time = HLF.merge_dates_post(record)
        if post_date_time:
            # remove microseconds from datetime
            self.key["timeofday"] = post_date_time.time.replace(microsecond=0)


class HourofdayBase(hll.HllBase):
    """Extends Base Class"""

    NAME = hll.HllBaseRef(facet=FACET, base="hourofday")

    def __init__(self, record: lbsn.Post = None):
        super().__init__()
        self.key["hourofday"] = None
        if record is None:
            # init empty
            return
        post_date_time = HLF.merge_dates_post(record)
        if post_date_time:
            # remove seconds and microseconds from datetime
            self.key["hourofday"] = post_date_time.time.replace(second=0, microsecond=0)


class DayofweekBase(hll.HllBase):
    """Extends Base Class"""

    NAME = hll.HllBaseRef(facet=FACET, base="dayofweek")

    def __init__(self, record: lbsn.Post = None):
        super().__init__()
        self.key["weekday"] = None
        if record is None:
            # init empty
            return
        post_date_time = HLF.merge_dates_post(record)
        if post_date_time:
            self.key["weekday"] = post_date_time.weekday


class DayofmonthBase(hll.HllBase):
    """Extends Base Class"""

    NAME = hll.HllBaseRef(facet=FACET, base="dayofmonth")

    def __init__(self, record: lbsn.Post = None):
        super().__init__()
        self.key["dayofmonth"] = None
        if record is None:
            # init empty
            return
        post_date_time = HLF.merge_dates_post(record)
        if post_date_time:
            self.key["dayofmonth"] = post_date_time.day


class DayofyearBase(hll.HllBase):
    """Extends Base Class"""

    NAME = hll.HllBaseRef(facet=FACET, base="dayofyear")

    def __init__(self, record: lbsn.Post = None):
        super().__init__()
        self.key["month"] = None
        self.key["day"] = None
        if record is None:
            # init empty
            return
        post_date_time = HLF.merge_dates_post(record)
        if post_date_time:
            self.key["month"] = post_date_time.month
            self.key["day"] = post_date_time.day


class MonthofyearBase(hll.HllBase):
    """Extends Base Class"""

    NAME = hll.HllBaseRef(facet=FACET, base="monthofyear")

    def __init__(self, record: lbsn.Post = None):
        super().__init__()
        self.key["monthofyear"] = None
        if record is None:
            # init empty
            return
        post_date_time = HLF.merge_dates_post(record)
        if post_date_time:
            self.key["monthofyear"] = post_date_time.month


class MonthHashtagBase(hll.HllBase):
    """Composite Base (c-base) that extends from HLL base Class

    Note: To distinguish c-bases which are composite bases combining
    aspects from multiple facets, they're termed with a leading underscore
    """

    NAME = hll.HllBaseRef(facet=FACET, base="_month_hashtag")

    def __init__(self, record: lbsn.Post = None, hashtag: str = None):
        super().__init__()
        self.key["year"] = None
        self.key["month"] = None
        self.key["hashtag"] = None
        if hashtag is None:
            # init empty
            return
        self.key["hashtag"] = hashtag.lower()
        if record is None:
            # init empty
            return
        if isinstance(record, lbsn.Post):
            post_date_time = HLF.merge_dates_post(record)
            if post_date_time:
                date = post_date_time.date()
                self.key["year"] = date.year
                self.key["month"] = date.month
        else:
            raise ValueError(
                "Parsing of MonthHashtagBase only supported " "from lbsn.Post"
            )


class MonthLatLngBase(hll.HllBase):
    """Composite Base (c-base) that extends from HLL base Class

    Note: To distinguish c-bases which are composite bases combining
    aspects from multiple facets, they're termed with a leading underscore
    """

    NAME = hll.HllBaseRef(facet=FACET, base="_month_latlng")

    def __init__(self, record: lbsn.Post = None):
        super().__init__()
        self.key["year"] = None
        self.key["month"] = None
        self.key["latitude"] = None
        self.key["longitude"] = None
        self.attrs["latlng_geom"] = None
        if record is None:
            # init empty
            return
        if isinstance(record, lbsn.Post):
            post_date_time = HLF.merge_dates_post(record)
            if post_date_time is None:
                return
            date = post_date_time.date()
            self.key["year"] = date.year
            self.key["month"] = date.month

            coordinates_geom = record.post_latlng
            coordinates = HF.get_coordinates_from_ewkt(coordinates_geom)
            self.key["latitude"] = coordinates.lat
            self.key["longitude"] = coordinates.lng
            # additional (optional) attributes
            # formatted ready for sql upsert
            self.attrs["latlng_geom"] = HF.return_ewkb_from_geotext(coordinates_geom)
        else:
            raise ValueError(
                "Parsing of MonthLatLngBase only supported " "from lbsn.Post"
            )


class MonthHashtagLatLngBase(hll.HllBase):
    """Composite Base (c-base) that extends from hll.HllBase Class

    Note: To distinguish c-bases which are composite bases combining
    aspects from multiple facets, they're termed with a leading underscore
    """

    NAME = hll.HllBaseRef(facet=FACET, base="_month_hashtag_latlng")

    def __init__(self, record: lbsn.Post = None, hashtag: str = None):
        super().__init__()
        self.key["year"] = None
        self.key["month"] = None
        self.key["hashtag"] = None
        self.key["latitude"] = None
        self.key["longitude"] = None
        self.attrs["latlng_geom"] = None
        if hashtag is None:
            # init empty
            return
        self.key["hashtag"] = hashtag.lower()

        if record is None:
            # init empty
            return

        if not isinstance(record, lbsn.Post):
            raise ValueError(
                "Parsing of MonthHashtagLatLngBase only supported " "from lbsn.Post"
            )
        post_date_time = HLF.merge_dates_post(record)
        if post_date_time:
            date = post_date_time.date()
            self.key["year"] = date.year
            self.key["month"] = date.month
        coordinates_geom = record.post_latlng
        coordinates = HF.get_coordinates_from_ewkt(coordinates_geom)
        self.key["latitude"] = coordinates.lat
        self.key["longitude"] = coordinates.lng
        # additional (optional) attributes
        # formatted ready for sql upsert
        self.attrs["latlng_geom"] = HF.return_ewkb_from_geotext(coordinates_geom)
